////////////////////////////////////////////////////////////////////////////////
///                             ir_machine.c                                 ///
///    IR remote control decoder with automatic protocol detection           ///
///                                                                          ///
/// Copyright 2017 Jaakko Kairus                                             ///
/// jaakko@kair.us                                                           ///
///                                                                          ///
/// You may freely use and modify this software for personal use.            ///
/// Commercial use of this software or parts of it is prohibited             ///
/// without written permission from author.                                  ///
/// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY              ///
///                                                                          ///
/// For more information about this project see project web page at          ///
/// http://kair.us/projects/clock/v4/index.html                              ///
///                                                                          ///
/// Developed with CCS compiler version 5.076                                ///
///                                                                          ///
////////////////////////////////////////////////////////////////////////////////

#ifndef _PCH
#byte STATUS = 0x03
#else
#byte STATUS = 0x0FD8
#endif
#bit CARRY = STATUS.0 

#include "ir_defines.h"

typedef enum {
        IR_IDLE,
        IR_START_PULSE,
        IR_START_PAUSE,
        IR_NEC_REPETITION,
        IR_NEC_DATA_PULSE,
        IR_NEC_DATA_PAUSE,
		IR_NEC_DATA_FINISHED,
		IR_NEC_DECODE,
		IR_NEC42_DECODE,
		IR_ERROR,
		IR_JVC_DECODE,
		IR_RC5_HALF_OF_BIT,
		IR_RC5_DATA_1,
		IR_RC5_DATA_0,
		IR_RC5_DECODE,
		IR_SIRC_DATA_PULSE,
		IR_SIRC_DATA_PAUSE,
		IR_SIRC_DATA_FINISHED,
		IR_SIRC_DECODE,
		IR_JAPAN_START_PAUSE,
		IR_KASEIKYO_DECODE,
		IR_MATSUSHITA_DECODE
} IR_STATES;
static IR_STATES IR_state = IR_IDLE;

static unsigned int16 actcycles = 0;		// How many timer cycles since last keypress, is the repetition still valid? 
static unsigned int8 repeats=0;
static unsigned int32 ir_receive_buffer;
static unsigned int16 ir_receive_buffer_16;
static unsigned int8 command_high;
static unsigned int8 command_low;
static unsigned int8 address_high;
static unsigned int8 address_low;
static unsigned int8 BIT_TIMER;
static unsigned int8 new_packet = 0;
static unsigned int8 is_samsung;
static unsigned int8 mark_bit;
static unsigned int8 bits_received;
static unsigned int8 bit_threshold;

void ir_machine(void);
int1 getIrPacket(IR_PACKET *completedPacket);

void ir_machine(void) {
	switch (IR_state) {
		case IR_IDLE:
			//if (new_packet)	// Do not receive new ir before previous 
			//	return FALSE;	// has been procssed
			if (!new_packet) {	// Globally clear address and command high bytes since they are not needed on all protocols
				command_high = 0;
				address_high = 0;
			}	

			if (actcycles!=0)
				actcycles--;
			else
				repeats=0;
			
			if (input(IRin)) // Input high (inactive)?
				return;
			BIT_TIMER=0;
			IR_state=IR_START_PULSE;
			mark_bit=0;
			bits_received=0;
			return;
		case IR_START_PULSE:
			BIT_TIMER++;
			if (!(input(IRin))) { // Pulse active?
				if (BIT_TIMER>NEC_START_PULSE_MAX) { // Longer than longest supported protocol (9 ms + 10%) low?
					IR_state=IR_ERROR;
				}
			}
			else {
				if (BIT_TIMER>NEC_START_PULSE_MIN && BIT_TIMER<NEC_START_PULSE_MAX) {  // 8.4ms -10% to 9ms + 10% low?
					IR_state=IR_START_PAUSE;		  // NEC or JVC
					is_samsung=0;
					BIT_TIMER=0;
				}
				else if (BIT_TIMER>SAMSUNG_START_PULSE_MIN && BIT_TIMER<SAMSUNG_START_PULSE_MAX) {  // 4.5ms +/- 10% low?
					IR_state=IR_START_PAUSE;		  // Samsung
					is_samsung=1;
					BIT_TIMER=0;
				}
				else if (BIT_TIMER>JAPAN_START_PULSE_MIN && BIT_TIMER<JAPAN_START_PULSE_MAX) { // 3.45 ms +/- 10% low?
					IR_state=IR_JAPAN_START_PAUSE;		 // Matsushita or Kaseikyo
					BIT_TIMER=0;
				}	
				else if (BIT_TIMER>SIRC_START_PULSE_MIN && BIT_TIMER<SIRC_START_PULSE_MAX) { // 2.4m +/- 10% low?
					IR_state=IR_SIRC_DATA_PAUSE;
					BIT_TIMER=0;
				}	
				else if (BIT_TIMER>RC5X_START_PULSE_MIN && BIT_TIMER<RC5X_START_PULSE_MAX) {	// RC-5 extended
					BIT_TIMER=0;
					shift_left(&ir_receive_buffer, 4, 0);
					bits_received++;
					IR_state=IR_RC5_HALF_OF_BIT;
				}	
				else if (BIT_TIMER>RC5_START_PULSE_MIN && BIT_TIMER<RC5_START_PULSE_MAX) {
					IR_state=IR_RC5_HALF_OF_BIT;	// RC-5
				}	
				else if (BIT_TIMER>JVC_REPEAT_PULSE_MIN && BIT_TIMER<JVC_REPEAT_PULSE_MAX) {
					IR_state = IR_NEC_DATA_PAUSE;	// JVC repeat. Go to NEC routine and determine first data bit
					is_samsung=0;
				}	
				else {
					IR_state=IR_IDLE;	  // too short, go back to idle
				}
			}
			return;
		case IR_START_PAUSE:			// Measure pause after first AGC pulse
			BIT_TIMER++;
			if (input(IRin)) {			// Input high?
				if (BIT_TIMER>NEC_START_PAUSE_MAX) {		// Longer than 4.5ms + 10% high?
					IR_state=IR_IDLE;		// Yes, not good! Start from state 0.
				}
			}
			else {						// Input is low.. 
				if (BIT_TIMER>NEC_REPETITION_THRESHOLD) {		// Let's see if the gap was long (>3.35ms)
					IR_state=IR_NEC_DATA_PULSE;			// It was long, go to state 4 and decode message
					bit_threshold = NEC_BIT_THRESHOLD;
				}
				else {
					IR_state=IR_NEC_REPETITION;			// It was a short one! It's a repetition!
				}
				BIT_TIMER=0;
			}
			return;
		case IR_NEC_REPETITION:
			IR_state=IR_ERROR;
			if (actcycles!=0) {   // Is the repetition valid?
				actcycles=NEC_REPEAT_INTERVAL;
				if (repeats<5)
					repeats++;
				if (repeats==0 || repeats==5)
					new_packet = NEC_PROTOCOL;
			}
			return;
		case IR_NEC_DATA_PULSE:
			BIT_TIMER++;
			if (!(input(IRin))) {  // Input still low?
//				if (BIT_TIMER>NEC_PULSE_MAX) { // See if max pulse width reached
				if (BIT_TIMER>bit_threshold/2) { // See if max pulse width reached
					IR_state=IR_ERROR; 			// Max width reached, goto error state to wait it gets high
				}
			}
			else {
				IR_state=IR_NEC_DATA_PAUSE;		// Input is now high
			}
			return;	
		case IR_NEC_DATA_PAUSE:
			BIT_TIMER++;
			if (BIT_TIMER < bit_threshold*2) { // Should we keep waiting?
				if (input(IRin)) { // Yes. Is the input still high?
				}
				else {
					if (bits_received<32)
						shift_right(&ir_receive_buffer, 4, (BIT_TIMER>bit_threshold));	// Use 1.5 bit time as threshold
					else
						shift_right(&ir_receive_buffer_16, 2, (BIT_TIMER>bit_threshold));	// Use 1.5 bit time as threshold
					bits_received++;
					IR_state = IR_NEC_DATA_PULSE;
					BIT_TIMER = 0;
				}
			}
			else {  // Data has been ended.
				if (bits_received == 16)
					IR_state = IR_JVC_DECODE; // Check if we have valid JVC code
				else if (bits_received == 24)
					IR_state = IR_MATSUSHITA_DECODE;
				else if (bits_received == 32)
					IR_state = IR_NEC_DECODE;
				else if (bits_received == 42)
					IR_state = IR_NEC42_DECODE;
				else if (bits_received == 48)
					IR_state = IR_KASEIKYO_DECODE;
				else
					IR_state = IR_ERROR;
			}
			return;
		case IR_NEC_DECODE:
			address_low = ir_receive_buffer;
			address_high = ir_receive_buffer>>8;
			command_low = ir_receive_buffer>>16;
			command_high = ir_receive_buffer>>24;
			
			IR_state=IR_IDLE;

			actcycles=NEC_REPEAT_INTERVAL;		

			repeats++;

			if (repeats==1 || repeats==5) {
				if (is_samsung)
					new_packet = SAMSUNG_PROTOCOL;
				else
					new_packet = NEC_PROTOCOL;
			}		
			if (repeats==5)
				repeats=4;
			return;
		case IR_NEC42_DECODE:
			address_low = ir_receive_buffer;				// address 8 lower bits
			address_high = ir_receive_buffer>>8 & 0x1f;	// address 5 upper bits
			command_low = (ir_receive_buffer>>26) + (ir_receive_buffer_16&0xc0);	// command
			command_high = ir_receive_buffer_16>>8;		// !command
			
			if ((command_high ^ command_low) != 0xff) {	// verify command with its complement
				IR_state=IR_ERROR;
				return;
			}	
				
			IR_state=IR_IDLE;

			actcycles=NEC_REPEAT_INTERVAL;		

			repeats++;

			if (repeats==1 || repeats==5) {
				new_packet = NEC42_PROTOCOL;
			}		
			if (repeats==5)
				repeats=4;
			return;
		case IR_ERROR:
			if (input(IRin)) { // Return to idle state when receiver goes idle
				IR_state=IR_IDLE;
			}
			return;
		case IR_JVC_DECODE:
			command_low = ir_receive_buffer>>24;
			address_low = ir_receive_buffer>>16;
			
			IR_state=IR_IDLE;
			actcycles=JVC_REPEAT_INTERVAL;
			repeats++;

			if (repeats==1 || repeats==8)
				new_packet = JVC_PROTOCOL;	// JVC protocol
			if (repeats==9)
				repeats=7;
			return;

		case IR_RC5_HALF_OF_BIT:
			if (bits_received==13) {
				IR_state = IR_RC5_DECODE;
				return;
			}	
			BIT_TIMER++;				// Wait until center of first half of bit
			if (BIT_TIMER<RC5_1B5)
				return;	// Not yet!
			if (input(IRin))	// If input is high, go to state 3
				IR_state=IR_RC5_DATA_1;
			else				// If it's low, go to state 4
				IR_state=IR_RC5_DATA_0;
			return;
		case IR_RC5_DATA_1:	// was 7839 bytes with asm routines on this and following two states
			BIT_TIMER++;	// now 7901 with C routines, but also handles RC5 extended protcol
			if (BIT_TIMER>2*RC5_1B5) {	// 1,5 times longer than should
				IR_state = IR_ERROR;	// Took too long
				return;
			}	
			if (input(IRin))	// Is it still high?
				return;			// We will wait...
			
			shift_left(&ir_receive_buffer, 4, 1);
			bits_received++;
			BIT_TIMER = 0;
			IR_state = IR_RC5_HALF_OF_BIT;				
			return;
		case IR_RC5_DATA_0:
			BIT_TIMER++;
			if (BIT_TIMER>2*RC5_1B5) {	// 1,5 times longer than should
				IR_state = IR_ERROR;	// Took too long
				return;
			}	
			if (!(input(IRin)))	// Is it still low?
				return;			// We will wait...

			shift_left(&ir_receive_buffer, 4, 0);
			bits_received++;
			BIT_TIMER = 0;
			IR_state = IR_RC5_HALF_OF_BIT;
			return;			
		case IR_RC5_DECODE:
			address_low=(ir_receive_buffer>>6)&0x1f;	// 5-bit address
			command_low=ir_receive_buffer&0x3f;	// 6-bit command
			if (!(ir_receive_buffer&0x1000))	// 7th command bit
				command_low+=0x40;
			IR_state = IR_ERROR;		
			actcycles=RC5_REPEAT_INTERVAL;		
			repeats++;
			if (repeats==1 || repeats==5) {
				new_packet = RC5_PROTOCOL;
			}		
			if (repeats==5)
				repeats=4;
			return;
		case IR_SIRC_DATA_PULSE:
			BIT_TIMER++;
			if (BIT_TIMER < SIRC_BIT_THRESHOLD*2) { // Should we keep waiting?
				if (!(input(IRin))) { // Yes. Is the input still low?
					return;  // It is - we'll wait..
				}
				else {
					shift_right(&ir_receive_buffer, 4, (BIT_TIMER>SIRC_BIT_THRESHOLD));	// Use 1.5 bit time as threshold
					bits_received++;
					IR_state = IR_SIRC_DATA_PAUSE;
					BIT_TIMER = 0;
               		
					return;
				}
			}
			else {  // Pulse is too long. It is not SIRC
				IR_state = IR_ERROR;
				return;
			}
		case IR_SIRC_DATA_PAUSE:
			BIT_TIMER++;
			if (input(IRin)) {  // Input still high?
				if (BIT_TIMER>SIRC_BIT_THRESHOLD) { // See if max pause width reached
					if (bits_received == 12 || bits_received == 15 || bits_received == 20)	// It can also be actual message end...
						IR_state = IR_SIRC_DATA_FINISHED;
					else
						IR_state=IR_ERROR; 			// Max width reached, goto error state to wait it gets high
				}
			}
			else {
				IR_state=IR_SIRC_DATA_PULSE;		// Input is now low
				BIT_TIMER = 0;
			}
			return;	
		case IR_SIRC_DATA_FINISHED:	// Additional wait that nothing is following in 4*900us = 3.6 ms
			BIT_TIMER++;
			if (input(IRin)) {
				if (BIT_TIMER>SIRC_BIT_THRESHOLD*4) {
					IR_state = IR_SIRC_DECODE;
				}
			}
			else {
				IR_state=IR_ERROR;
			}
			return;
		case IR_SIRC_DECODE:
			address_high=ir_receive_buffer>>(47-bits_received);	// 5-bit extra
			address_low=ir_receive_buffer>>(39-bits_received);	// 8-bit address
			command_low=(ir_receive_buffer>>(32-bits_received))&0x7f;	// 7-bit data

			IR_state=IR_IDLE;		// We're done! Let's start from beginning
			actcycles=SIRC_REPEAT_INTERVAL;

			repeats++;

			if (repeats==1 || repeats==10)
				new_packet = SIRC_PROTOCOL;
			if (repeats==11)
				repeats=9;

			return;
		case IR_JAPAN_START_PAUSE:			// Measure pause after first AGC pulse
			BIT_TIMER++;
			if (input(IRin)) {			// Input high?
				if (BIT_TIMER>JAPAN_START_PAUSE_MAX) {		// Longer than 3.5ms + 10% high?
					IR_state=IR_IDLE;		// Yes, not good! Start from state 0.
				}
			}
			else {						// Input is low.. 
				if (BIT_TIMER>MATSUSHITA_PAUSE_THRESHOLD) {		// Let's see if the gap was long (>2.6ms)
					IR_state=IR_NEC_DATA_PULSE;			// It is Matsushita or Technics
					bit_threshold = MATSUSHITA_BIT_THRESHOLD;
				}
				else {
					IR_state = IR_NEC_DATA_PULSE;	// It was a short one. It is Kaseikyo
					bit_threshold = KASEIKYO_BIT_THRESHOLD;
				}
				BIT_TIMER=0;
			}
			return;
		case IR_KASEIKYO_DECODE:
			address_low = ir_receive_buffer>>8;
			address_high = ir_receive_buffer;
			command_low = ir_receive_buffer_16;
			command_high = ir_receive_buffer>>24;
			
			IR_state=IR_IDLE;

			actcycles=KASEIKYO_REPEAT_INTERVAL;		

			repeats++;

			if (repeats==1 || repeats==5) {
				new_packet = KASEIKYO_PROTOCOL;
			}		
			if (repeats==5)
				repeats=4;
			return;
		case IR_MATSUSHITA_DECODE:
			address_low = ir_receive_buffer & 0x3f;				// 6-bit address
			command_low = (ir_receive_buffer>>6) & 0x3f;			// 6-bit command
			// Inverted byte checking is disabled below because not enough memory currently.
/*			address_high = (ir_receive_buffer>>12) & 0x3f;		// 6-bit inverted address
			command_high = ir_receive_buffer>>18 & 0x3f;		// 6-bit inverted command
			
			if ((command_high ^ command_low) != 0xff) {			// verify command with its complement
				IR_state=IR_ERROR;
				return;
			}
*/			
			IR_state=IR_IDLE;

			actcycles=MATSUSHITA_REPEAT_INTERVAL;		

			repeats++;

			if (repeats==1 || repeats==5) {
				new_packet = MATSUSHITA_PROTOCOL;
			}		
			if (repeats==5)
				repeats=4;
			return;

//		default:	// Should not get here but let's get out anyway
//			IR_state = IR_IDLE;
//			return;
	}
}





int1 getIrPacket(IR_PACKET *completedPacket) {
	if (new_packet) {
		completedPacket->protocol = new_packet;
		completedPacket->address = address_high*256+address_low;
		completedPacket->command = command_high*256+command_low;
		completedPacket->flags = 0x00;
		new_packet=0;
		return TRUE;
	}
	else
		return FALSE;	
}	