////////////////////////////////////////////////////////////////////////////////
///                               PCF2129.C                                  ///
///                     Driver for Real Time Clock                           ///
/// Copyright 2017-2018 Jaakko Kairus                                        ///
/// jaakko@kair.us                                                           ///
///                                                                          ///
/// You may freely use and modify this software for personal use.            ///
/// Commercial use of this software or parts of it is prohibited             ///
/// without written permission from author.                                  ///
/// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY              ///
///                                                                          ///
/// For more information about this project see project web page at          ///
/// http://kair.us/projects/clock/v4/index.html                              ///
///                                                                          ///
/// Developed with CCS compiler version 5.077                                ///
///                                                                          ///
/// Recommended to use combined set_date_and_time and get_date_and_time      ///
/// functions to prevent possible time/date corruption.                      ///
///                                                                          ///
/// PCF2129 is initialized to PWRMNG[2:0] = 000                              ///
/// battery switch-over function is enabled in standard mode;                ///
/// battery low detection function is enabled                                ///
/// The standard mode consumes 2.15 A which is about double compared to     ///
/// direct switching mode, but since the Vbat and Vdd have almost same level ///
/// there is risk to use Vbat if battery is full, causing I2C to be disabled ///
/// With the 2.15 A consumption, a typical 190 mAh CR2032 battery still     ///
/// lasts over 10 years.                                                     ///
////////////////////////////////////////////////////////////////////////////////

// #define RTC_SDA  PIN_C4
// #define RTC_SCL  PIN_C3
#define PCF2129_WRITE_ADDR	0xa2
#define PCF2129_READ_ADDR	0xa3

#define PCF2129_SEC_ALARM_REG 0x0a
#define PCF2129_MIN_ALARM_REG 0x0b
#define PCF2129_HOUR_ALARM_REG 0x0c

// #use i2c(master, sda=RTC_SDA, scl=RTC_SCL)

int8 pcf2129_init();
void pcf2129_set_time_and_date(int8 sec, int8 min, int8 hr, int8 day, int8 dow, int8 mth, int8 year);
void pcf2129_set_byte(int8 data, int8 address);
void pcf2129_get_byte(int8* data, int8 address);
#separate void pcf2129_get_struct_tm(struct_tm* utc);
int8 bin2bcd(int8 binary_value);
int8 bcd2bin(int8 bcd_value);

int8 pcf2129_init() {
	int8 osf=1,i=0;
	
	while (osf) {			// Wait for ack from PCF before continuing with init to ensure the oscillator has started incase of no battery installed
		i2c_start();
		osf = i2c_write(PCF2129_WRITE_ADDR);
		i2c_stop();
		delay_ms(10);
		i++;
		if (i>250)		// no ack from PCF in 2.5 seconds
			return 2;	// RTC chip not found!
	}	
	
	pcf2129_set_byte(0x00, 0x00); // Clear PORO
	pcf2129_set_byte(0b00000110, 0x0f); // 0x0f is clkout control reg. Set clkout to 1 Hz
	pcf2129_get_byte(&osf, 0x03);
	if (osf&0x80) {
		#ifdef TIME_T_USES_2010
		pcf2129_set_time_and_date(0,0,0,1,1,1,18);	// zero clock registers. 1.1.2018 was monday (1)
		#else
		pcf2129_set_time_and_date(0,0,0,1,6,1,0);	// zero clock registers. 1.1.2000 was saturday (6)
		#endif
		pcf2129_set_byte(0b00000110, 0x0f); // toggle OTPR bit of clockout control reg
		pcf2129_set_byte(0b00100110, 0x0f); // recommended is set OTPR to 0 and then 1 to perform OTP refresh after first power on
		delay_ms(100);						// the OTP refresh takes less than 100 ms, ensure that nothing funny happens during that time
		return 1;		// OSF (oscillator stop) flag was set, time register contents zeroed
	}	
	else
		return 0;		// All OK
}

void pcf2129_set_time(int8 hr, int8 min, int8 sec)
{
  sec &= 0x7F;
  // hr &= 0x3F;

  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);            // I2C write address
  i2c_write(0x03);            // Start at REG 3 - Seconds
  i2c_write(bin2bcd(sec));      // REG 3
  i2c_write(bin2bcd(min));      // REG 4
  i2c_write(bin2bcd(hr));      // REG 5
  i2c_stop();
}

void pcf2129_set_date(int8 day, int8 mth, int8 year, int8 dow)
{
  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);            // I2C write address
  i2c_write(0x06);            // Start at REG 6 - Day 
  i2c_write(bin2bcd(day));      // REG 6
  i2c_write(bin2bcd(dow));      // REG 7
  i2c_write(bin2bcd(mth));      // REG 8
  i2c_write(bin2bcd(year));      // REG 9
  i2c_stop();
}

void pcf2129_set_time_and_date(int8 sec, int8 min, int8 hr, int8 day, int8 dow, int8 mth, int8 year)
{
  sec &= 0x7F;					// Mask MSB because it is OSF (oscillator stopped)
  // hr &= 0x3F;

  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);            // I2C write address
  i2c_write(0x03);            // Start at REG 3 - Seconds
  i2c_write(bin2bcd(sec));      // REG 3
  i2c_write(bin2bcd(min));      // REG 4
  i2c_write(bin2bcd(hr));      // REG 5
  i2c_write(bin2bcd(day));      // REG 6
  i2c_write(bin2bcd(dow));      // REG 7
  i2c_write(bin2bcd(mth));      // REG 8
  i2c_write(bin2bcd(year));      // REG 9
  i2c_stop();
}

void pcf2129_set_struct_tm(struct_tm* utc)
{
  utc->tm_sec &= 0x7F;					// Mask MSB because it is OSF (oscillator stopped)
  // hr &= 0x3F;

  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);            // I2C write address
  i2c_write(0x03);            // Start at REG 3 - Seconds
  i2c_write(bin2bcd(utc->tm_sec));      // REG 3
  i2c_write(bin2bcd(utc->tm_min));      // REG 4
  i2c_write(bin2bcd(utc->tm_hour));      // REG 5
  i2c_write(bin2bcd(utc->tm_mday));      // REG 6
  i2c_write(bin2bcd(utc->tm_wday));      // REG 7
  i2c_write(bin2bcd((utc->tm_mon)+1));      // REG 8
  i2c_write(bin2bcd((utc->tm_year)-100));      // REG 9
  i2c_stop();
}

void pcf2129_set_alarm(int8 hr, int8 min, int8 sec)
{
  sec &= 0x7F;
  hr &= 0x3F;

  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);            // I2C write address
  i2c_write(0x0A);            // Start at REG a - Alarm seconds
  i2c_write(bin2bcd(sec));      // REG 7
  i2c_write(bin2bcd(min));      // REG 8
  i2c_write(bin2bcd(hr));      // REG 9
  i2c_stop();
}

void pcf2129_get_date(int8 &day, int8 &mth, int8 &year, int8 &dow)
{
  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);
  i2c_write(0x06);            // Start at REG 6 - Day
  i2c_stop();				// Repeated start is not allowed with PCF2129
  i2c_start();
  i2c_write(PCF2129_READ_ADDR);
  day  = bcd2bin(i2c_read() & 0x3f);
  dow  = bcd2bin(i2c_read() & 0x07);
  mth  = bcd2bin(i2c_read() & 0x1f);
  year = bcd2bin(i2c_read(0));
  i2c_stop();
}

void pcf2129_get_time(int8 &hr, int8 &min, int8 &sec)
{
  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);
  i2c_write(0x03);            // Start at REG 3 - Seconds
  i2c_stop();				// Repeated start is not allowed with PCF2129
  i2c_start();
  i2c_write(PCF2129_READ_ADDR);
  sec = bcd2bin(i2c_read() & 0x7f);
  min = bcd2bin(i2c_read() & 0x7f);
  hr  = bcd2bin(i2c_read(0) & 0x3f);
  i2c_stop();
}

void pcf2129_get_time_and_date(int8 &sec, int8 &min, int8 &hr, int8 &day, int8 &dow, int8 &mth, int8 &year)
{
  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);
  i2c_write(0x03);            // Start at REG 3 - Seconds
  i2c_stop();				// Repeated start is not allowed with PCF2129
  i2c_start();
  i2c_write(PCF2129_READ_ADDR);
  sec = bcd2bin(i2c_read() & 0x7f);
  min = bcd2bin(i2c_read() & 0x7f);
  hr  = bcd2bin(i2c_read() & 0x3f);
  day  = bcd2bin(i2c_read() & 0x3f);
  dow  = bcd2bin(i2c_read() & 0x07);
  mth  = bcd2bin(i2c_read() & 0x1f);
  year = bcd2bin(i2c_read(0));
  i2c_stop();
}

void pcf2129_get_struct_tm(struct_tm* utc)
{
  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);
  i2c_write(0x03);            // Start at REG 3 - Seconds
  i2c_stop();				// Repeated start is not allowed with PCF2129
  i2c_start();
  i2c_write(PCF2129_READ_ADDR);
  utc->tm_sec = bcd2bin(i2c_read() & 0x7f);
  utc->tm_min = bcd2bin(i2c_read() & 0x7f);
  utc->tm_hour  = bcd2bin(i2c_read() & 0x3f);
  utc->tm_mday  = bcd2bin(i2c_read() & 0x3f);
  utc->tm_wday  = bcd2bin(i2c_read() & 0x07);
  utc->tm_mon  = bcd2bin(i2c_read() & 0x1f)-1;
  utc->tm_year = bcd2bin(i2c_read(0))+100;
  i2c_stop();
}


void pcf2129_get_alarm(int8 &hr, int8 &min, int8 &sec)
{
  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);
  i2c_write(0x0a);
  i2c_stop();				// Repeated start is not allowed with PCF2129
  i2c_start();
  i2c_write(PCF2129_READ_ADDR);
  sec = bcd2bin(i2c_read() & 0x7f);
  min = bcd2bin(i2c_read() & 0x7f);
  hr  = bcd2bin(i2c_read(0) & 0x3f);
  i2c_stop();
}

void pcf2129_set_byte(int8 data, int8 address )
{
  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);
  i2c_write(address);
  i2c_write(data);
  i2c_stop();
}

void pcf2129_get_byte(int8* data, int8 address)
{
  i2c_start();
  i2c_write(PCF2129_WRITE_ADDR);
  i2c_write(address);
  i2c_stop();				// Repeated start is not allowed with PCF2129
  i2c_start();
  i2c_write(PCF2129_READ_ADDR);
  *data  = i2c_read(0);
  i2c_stop();
}

int8 bin2bcd(int8 binary_value) {
	int8 temp;
	int8 retval;

	temp = binary_value;
	retval = 0;
	
	while (temp >= 10) {
		temp -= 10;			// Get the tens digit by doing multiple subtraction
		retval += 0x10;		// of 10 from the binary value.
	}
	
	retval += temp;			// Get the ones digit by adding the remainder.
	return(retval);
}


// Input range - 00 to 99.
int8 bcd2bin(int8 bcd_value)
{
  int8 temp;

  temp = bcd_value;
  // Shifting upper digit right by 1 is same as multiplying by 8.
  temp >>= 1;
  // Isolate the bits for the upper digit.
  temp &= 0x78;

  // Now return: (Tens * 8) + (Tens * 2) + Ones

  return(temp + (temp >> 2) + (bcd_value & 0x0f));
}


/*
//subroutine converting BCD to Binary.   
int8 bcd2bin(int8 bcd_value)
{
return((bcd_value >> 4) * 10 + (bcd_value & 0x0F));
}

//---------------------------------------
//subroutine converting Binary to BCD.
int8 bin2bcd(int8 binary_value)
{
return( (swap(binary_value/10)) & (binary_value % 10));
} 
*/
/*
unsigned int8 bin2bcd(unsigned int8 data)
{
   unsigned int8 nibh;
   unsigned int8 nibl;

   nibh=data/10;
   nibl=data-(nibh*10);

   return((nibh<<4)|nibl);
}
                                         
unsigned int8 bcd2bin(unsigned int8 data)
{                                                         
   unsigned int8 i;                

   i=data;                     
   data=(i>>4)*10;
   data=data+(i & 0x0F);       

   return data;
}
*/